"""flyswot command line"""
import typer

from flyswot import cli_inference
from flyswot import models

app = typer.Typer()

app.add_typer(
    cli_inference.app, name="predict", help="flyswot commands for making predictions"
)
app.add_typer(
    models.app, name="model", help="flyswot commands for interacting with models"
)

typer_click_object = typer.main.get_command(app)


def main() -> None:  # pragma: no cover
    """flyswot command line application"""
    app()



"""Inference functionality"""
import csv
import mimetypes
import re
import string
import time
from collections import defaultdict
from collections import OrderedDict
from dataclasses import asdict
from datetime import datetime
from datetime import timedelta
from functools import singledispatch
from pathlib import Path
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import PIL
import typer
from rich import print
from rich.columns import Columns
from rich.layout import Layout
from rich.markdown import Markdown
from rich.panel import Panel
from rich.progress import Progress
from rich.table import Table
from rich.text import Text
from toolz import itertoolz
from toolz.dicttoolz import merge
from transformers import AutoFeatureExtractor
from transformers import AutoModelForImageClassification
from transformers import pipeline

from flyswot import core
from flyswot import models
from flyswot.console import console
from flyswot.inference import InferenceSession
from flyswot.inference import MultiLabelImagePredictionItem
from flyswot.inference import MultiPredictionBatch
from flyswot.inference import PredictionBatch
from flyswot.logo import flyswot_logo

app = typer.Typer()


image_extensions = {k for k, v in mimetypes.types_map.items() if v.startswith("image/")}


@app.command()
def predict_image(
    image: Path = typer.Argument(..., readable=True, resolve_path=True)
) -> None:
    """Predict a single image"""
    pass  # pragma: no cover


def check_files(files: List, pattern: str, directory: Path) -> None:
    """Check if files exist and raises error if not"""
    if not files:
        print(
            f"Didn't find any files maching {pattern} in {directory}, please check the inputs to flyswot"
        )
        raise typer.Exit(code=1)


def try_predict_batch(batch, inference_session, bs):
    """try and predict a batch of files"""
    bad_batch = False
    try:
        batch_predictions = inference_session.predict_batch(batch, bs)
        return batch_predictions, bad_batch
    except PIL.UnidentifiedImageError:
        bad_batch = True
        return batch, bad_batch
    except ValueError:
        bad_batch = True
        return batch, bad_batch


def predict_files(
    files: List[Path], inference_session, bs, csv_fname
) -> Tuple[set, int]:
    """Predict files"""
    with Progress() as progress:
        total_progress = progress.add_task("prediction progress", total=len(files))
        images_checked = 0
        bad_batch_files = []
        for i, batch in enumerate(itertoolz.partition_all(bs, files)):
            batch_predictions, bad_batch = try_predict_batch(
                batch, inference_session, bs
            )
            if bad_batch:
                bad_batch_files.append(batch)
            if i == 0 and not bad_batch:
                create_csv_header(batch_predictions, csv_fname)
            if not bad_batch:
                write_batch_preds_to_csv(batch_predictions, csv_fname)
            progress.update(total_progress, advance=len(batch))
            images_checked += len(batch)
        corrupt_images = set()
        if bad_batch_files:
            for batch in bad_batch_files:
                for file in batch:
                    try:
                        batch_predictions = inference_session.predict_batch([file], bs)
                        write_batch_preds_to_csv(batch_predictions, csv_fname)
                    except PIL.UnidentifiedImageError:
                        corrupt_images.add(file)
        return corrupt_images, images_checked


@app.command(name="directory")
def predict_directory(
    directory: Path = typer.Argument(
        ...,
        readable=True,
        resolve_path=True,
        help="Directory to start searching for images from",
    ),
    csv_save_dir: Path = typer.Argument(
        ...,
        writable=True,
        resolve_path=True,
        help="Directory used to store the csv report",
    ),
    model_id: str = typer.Option(
        "flyswot/convnext-tiny-224_flyswot",
        help="The model flyswot should use for making predictions",
    ),
    pattern: str = typer.Option(None, help="Pattern used to filter image filenames"),
    bs: int = typer.Option(16, help="Batch Size"),
    image_formats: List[str] = typer.Option(
        default=[".tif"],
        help="Image format(s) to check",
    ),
):
    """Predicts against all images stored under DIRECTORY which match PATTERN in the filename.

    By default searches for filenames containing 'fs'.

    Creates a CSV report saved to `csv_save_dir`
    """
    start_time = time.perf_counter()
    huggingfaceinference = HuggingFaceInferenceSession(model=model_id)
    files = sorted(
        itertoolz.concat(
            core.get_image_files_from_pattern(directory, pattern, image_format)
            for image_format in image_formats
        )
    )
    check_files(files, pattern, directory)
    if not pattern:
        pattern = "any pattern"
    print(
        f"Found {len(files)} files matching {pattern} in {directory} with extension(s) {image_formats}"
    )
    csv_fname = create_csv_fname(csv_save_dir)
    corrupt_images, images_checked = predict_files(
        files, inference_session=huggingfaceinference, bs=bs, csv_fname=csv_fname
    )
    if corrupt_images:
        print(corrupt_images)
    delta = timedelta(seconds=time.perf_counter() - start_time)
    print_inference_summary(
        str(delta),
        pattern,
        directory,
        csv_fname,
        image_formats,
        images_checked,
        model_id,
    )


def print_inference_summary(
    time_delta: str,
    pattern: str,
    directory: Path,
    csv_fname: Path,
    image_format: Union[List[str], str],
    matched_file_count: int,
    model_id: Optional[str] = None,
):
    """prints summary report"""
    print(flyswot_logo())
    if model_id:
        print(Panel(Columns([models.hub_model_link(model_id)]), title="Model Info"))

    print(
        Panel(
            Text(
                f"CSV report file at: {csv_fname.as_uri()}",
            ),
            title=" :clipboard: CSV report :clipboard:",
        )
    )
    print(
        Panel(
            Text(f"{time_delta}", justify="center", style="bold green"),
            title=":stopwatch: Time taken to run :stopwatch:",
        )
    )
    print(
        create_file_summary_markdown(
            pattern, matched_file_count, directory, image_format
        )
    )
    inference_summary_columns = get_inference_table_columns(csv_fname)
    print(Panel(Columns(inference_summary_columns), title="Prediction Summary"))


def create_file_summary_markdown(
    pattern: str,
    matched_file_count: int,
    directory: Path,
    image_formats: Union[List[str], str],
) -> Panel:
    """creates Markdown summary containing number of files checked by flyswot vs total images files under directory"""
    return Panel(
        Markdown(
            f"""
    - flyswot searched for image files by matching the patern *{pattern}* with extension(s) {image_formats}
    - flyswot search inside: `{directory}`
    - There were **{matched_file_count}** files matching the {pattern}* pattern which flyswot checked
    """
        ),
        title=":file_folder: files checked :file_folder:",
    )


def get_inference_table_columns(csv_fname: Path) -> List[Table]:
    """print_inference_summary from `fname`"""
    labels_to_print = labels_from_csv(csv_fname)
    return [
        print_table(labels, f"Prediction summary {i+1}", print=False)
        for i, labels in enumerate(labels_to_print)
    ]


label_regex = re.compile(r"prediction_label_\D_0")


def labels_from_csv(fname: Path) -> List[List[str]]:
    """Gets top labels from csv `fname`"""
    columns = defaultdict(list)
    with open(fname, "r") as f:
        reader = csv.DictReader(f)
        for row in reader:
            for (k, v) in row.items():
                columns[k].append(v)
    return [columns[k] for k in columns if label_regex.match(k)]


def print_table(
    decoded: list, header: str = "Prediction summary", print: bool = True
) -> Table:
    """Prints table summary of predicted labels"""
    table = Table(show_header=True, title=header)
    table.add_column(
        "Class",
    )
    table.add_column("Count")
    table.add_column("Percentage")
    total = len(decoded)
    frequencies = itertoolz.frequencies(decoded)
    for is_last_element, var in core.signal_last(frequencies.items()):
        key, value = var
        count = value
        percentage = round((count / total) * 100, 2)
        if is_last_element:
            table.add_row(key, str(count), f"{percentage}", end_section=True)
            table.add_row("Total", str(total), "")
        else:
            table.add_row(key, str(count), f"{percentage}")
    if print:
        console.print(table)
    return table


def make_layout():
    """Define the layout."""
    layout = Layout(name="root")
    layout.split(Layout(name="header", size=4), Layout(name="main"))
    layout["main"].split_column(
        Layout(name="info", size=4), Layout(name="body", ratio=2, minimum_size=60)
    )
    layout["info"].split_row(Layout(name="time"), Layout(name="files"))
    return layout


def create_csv_fname(csv_directory: Path) -> Path:
    """Creates a csv filename"""
    date_now = datetime.now()
    date_now = date_now.strftime("%Y_%m_%d_%H_%M")
    fname = Path(f"{date_now}.csv")
    return Path(csv_directory / fname)


@singledispatch
def create_csv_header(batch, csv_path: Path) -> None:
    """Print csv header"""
    raise NotImplementedError(f"Not implemented for type {batch}")


@create_csv_header.register
def _(batch: PredictionBatch, csv_path: Path):
    """Creates a header for csv `csv_path`"""
    with open(csv_path, mode="w", newline="") as csv_file:
        field_names = ["path", "directory", "prediction", "confidence"]
        writer = csv.DictWriter(csv_file, fieldnames=field_names)
        writer.writeheader()


@create_csv_header.register
def _(batch: MultiPredictionBatch, csv_path: Path, top_n: int = 2) -> None:
    item = batch.batch[0]
    pred = OrderedDict()
    pred["path"] = asdict(item)["path"]
    pred["directory"] = asdict(item)["path"].parent
    for i, _ in enumerate(item.predictions):
        for j in range(top_n):
            pred[f"prediction_label_{string.ascii_letters[i]}_{j}"] = ""
            pred[f"confidence_label_{string.ascii_letters[i]}_{j}"] = ""
    with open(csv_path, mode="w", newline="") as csv_file:
        field_names = list(pred.keys())
        writer = csv.DictWriter(csv_file, fieldnames=field_names)
        writer.writeheader()


@singledispatch
def write_batch_preds_to_csv(predictions, csv_fpath: Path) -> None:
    """print batch preds to csv"""
    raise NotImplementedError(f"Not implemented for type {predictions}")


@write_batch_preds_to_csv.register
def _(predictions: PredictionBatch, csv_fpath: Path) -> None:
    """Appends `predictions` batch to `csv_path`"""
    with open(csv_fpath, mode="a", newline="") as csv_file:
        field_names = ["path", "directory", "predicted_label", "confidence"]
        writer = csv.DictWriter(csv_file, fieldnames=field_names)
        for pred in predictions.batch:
            row = asdict(pred)
            row["directory"] = pred.path.parent
            writer.writerow(row)


@write_batch_preds_to_csv.register
def _(predictions: MultiPredictionBatch, csv_fpath: Path, top_n: int = 2) -> None:
    for pred in predictions.batch:
        row = OrderedDict()
        row["path"] = asdict(pred)["path"]
        row["directory"] = asdict(pred)["path"].parent
        for i, v in enumerate(pred.predictions):
            sorted_predictions = sorted(v.items(), reverse=True)
            for j in range(top_n):
                row[f"prediction_label_{i}_{j}"] = sorted_predictions[j][1]
                row[f"confidence_label_{i}_{j}"] = sorted_predictions[j][0]
        with open(csv_fpath, mode="a", newline="") as csv_file:
            field_names = list(row.keys())
            writer = csv.DictWriter(csv_file, fieldnames=field_names)
            writer.writerow(row)


class HuggingFaceInferenceSession(InferenceSession):
    "Huggingface inference session"

    def __init__(self, model: str):
        """Create Hugging Face Inference Session"""
        self.model = AutoModelForImageClassification.from_pretrained(model)
        self.feature_extractor = AutoFeatureExtractor.from_pretrained(model)
        self.session = pipeline(
            "image-classification",
            model=self.model,
            feature_extractor=self.feature_extractor,
        )

    def predict_image(self, image: Path) -> List[Dict[str, float]]:
        """Predict single Image."""
        return self.session(image, top_k=10)

    def predict_batch(self, batch: Iterable[Path], bs: int) -> MultiPredictionBatch:
        """Predict batch of images"""
        str_batch = [str(file) for file in batch]
        predictions: List[List[Dict[str, float]]] = self.session(
            str_batch, batch_size=bs, top_k=20
        )
        prediction_dicts = [self._process_prediction_dict(pred) for pred in predictions]
        all_pred = []
        for file, pred in zip(str_batch, prediction_dicts):
            item_pred = [pred]
            prediction = MultiLabelImagePredictionItem(Path(file), item_pred)
            all_pred.append(prediction)
        return MultiPredictionBatch(all_pred)

    def _process_prediction_dict(
        self, prediction_item: List[Dict[str, float]]
    ) -> Dict[float, str]:
        return merge(
            [
                {prediction["score"]: prediction["label"]}
                for prediction in prediction_item
            ]
        )


if __name__ == "__main__":
    app()  # pragma: no cover



"""Onnxruntime inference session."""
from pathlib import Path
from typing import Iterable
from typing import List
from typing import Union

import numpy as np
from PIL import Image

from flyswot import models
from flyswot.inference import ImagePredictionArgmaxItem
from flyswot.inference import InferenceSession
from flyswot.inference import MultiLabelImagePredictionItem
from flyswot.inference import MultiPredictionBatch
from flyswot.inference import PredictionBatch

try:  # pragma: no cover
    import onnxruntime as rt
except ImportError:  # pragma: no cover
    print("onnxruntime not found")


def softmax(x):
    """return softmax of `x`"""
    x = x.reshape(-1)
    e_x = np.exp(x - np.max(x))
    return e_x / e_x.sum(axis=0)


class OnnxInferenceSession(InferenceSession):  # pragma: no cover
    """onnx inference session"""

    def __init__(self, model: Path, vocab: Path):
        """Create onnx session"""
        self.model = model
        self.session = rt.InferenceSession(str(model))

        self.vocab = models.load_vocab(vocab)
        self.vocab_mappings = [dict(enumerate(v)) for v in self.vocab]

    def predict_image(
        self, image: Path
    ) -> Union[ImagePredictionArgmaxItem, MultiLabelImagePredictionItem]:
        """Predict a single image"""
        img = self._load_image(image)
        output_names = [o.name for o in self.session.get_outputs()]
        raw_result = self.session.run(output_names, {"image": img})
        if len(self.vocab) < 2:
            pred = self._postprocess(raw_result)
            arg_max = int(np.array(pred).argmax())
            predicted_label = self.vocab_mappings[0][arg_max]
            confidence = float(np.array(pred).max())
            return ImagePredictionArgmaxItem(image, predicted_label, confidence)
        else:
            prediction_dicts = []
            for vocab_map, pred in zip(self.vocab_mappings, raw_result):
                predictions = self._postprocess(pred)
                prediction_dict = {
                    float(prediction): vocab_map[i]
                    for i, prediction in enumerate(predictions)
                }
                prediction_dicts.append(prediction_dict)
        return MultiLabelImagePredictionItem(image, prediction_dicts)

    def _preprocess(self, input_data: np.ndarray) -> np.ndarray:
        # converts the input data into the float32 input for onnx
        img_data = input_data.astype("float32")

        # normalize
        mean_vec = np.array([0.485, 0.456, 0.406])
        stddev_vec = np.array([0.229, 0.224, 0.225])
        norm_img_data = np.zeros(img_data.shape).astype("float32")
        for i in range(img_data.shape[0]):
            norm_img_data[i, :, :] = (
                img_data[i, :, :] / 255 - mean_vec[i]
            ) / stddev_vec[i]

        # add batch channel
        norm_img_data = norm_img_data.reshape(1, 3, 512, 512).astype("float32")
        return norm_img_data

    def _load_image(self, file: Path) -> np.ndarray:
        """loads image and carries out preprocessing for inference"""
        image = Image.open(file, mode="r")
        image = image.resize((512, 512), Image.BILINEAR)
        image_data = np.array(image).transpose(2, 0, 1)
        return self._preprocess(image_data)

    def _postprocess(self, result: List):
        """process results from onnx session"""
        return softmax(np.array(result)).tolist()

    def predict_batch(
        self, batch: Iterable[Path], bs: int
    ) -> Union[PredictionBatch, MultiPredictionBatch]:
        """predicts a batch of images"""
        prediction_items = [self.predict_image(file) for file in batch]
        if len(self.vocab) < 2:
            return PredictionBatch(prediction_items)
        else:
            return MultiPredictionBatch(prediction_items)



"""Model Commands."""
import fnmatch
from dataclasses import dataclass
from pathlib import Path
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import requests
import typer
from huggingface_hub import hf_hub_url
from huggingface_hub import snapshot_download
from huggingface_hub.hf_api import ModelInfo
from rich import print
from rich.markdown import Markdown
from rich.table import Table
from toolz import itertoolz
from toolz import recipes

from flyswot.config import APP_NAME
from flyswot.config import MODEL_REPO_ID
from flyswot.console import console

app = typer.Typer()


@dataclass
class LocalModel:
    """A local model container"""

    hf_cache: Path

    def __post_init__(self):
        """Returns model parts contained under hf_cache"""
        self._get_model_parts(self.hf_cache)

    def _get_model_parts(self, hf_cache: Path):
        """Returns model path, vocab and metadata for a model"""
        model_files = Path(hf_cache).iterdir()
        for file in model_files:
            if fnmatch.fnmatch(file.name, "vocab.txt"):
                self.vocab = file
            if fnmatch.fnmatch(file.name, "README.md"):
                self.modelcard = file
            if fnmatch.fnmatch(file.name, "*.onnx") or fnmatch.fnmatch(
                file.name, "*.pkl"
            ):
                self.model = file


def ensure_model_dir(model_dir_path: Union[Path, None] = None) -> Path:
    """Checks for a local model dir and creates one if not found"""
    if not model_dir_path:
        app_dir = typer.get_app_dir(APP_NAME)
        model_dir: Path = Path(app_dir) / "models"
    else:
        model_dir = Path(model_dir_path) / "models"
    if not (model_dir.exists() and model_dir.is_dir()):
        print(f"Creating directory for storing models in {model_dir}...")
        try:
            model_dir.mkdir(parents=True)
        except PermissionError as e:  # pragma: no cover
            print(f"{model_dir} is not writeable: {e}")
            raise typer.Exit(code=1) from None
    print(f"Models stored in {model_dir}")
    return model_dir


@app.command()
def show_model_dir() -> None:
    """Print out the directory where models are stored"""
    ensure_model_dir()


@app.command(name="download")
def get_model(
    revision: Optional[str] = typer.Argument(None),
    model_dir: Path = typer.Argument(
        None,
        envvar="MODEL_DIR",
        help="Optionally specify a directory to store model files in",
    ),
    local_only=False,
) -> Path:  # pragma: no cover
    """Downloads models, defaults to the latest available model"""
    repo_id = MODEL_REPO_ID
    with console.status("Getting model", spinner="dots"):
        model = snapshot_download(
            repo_id, cache_dir=model_dir, revision=None, local_files_only=local_only
        )
    return Path(model)


def ensure_model(model_dir: Path) -> LocalModel:  # pragma: no cover
    """Checks for a local model and if not found downloads the latest available remote model"""
    if model := get_model(model_dir=model_dir):
        return LocalModel(model)
    print("Not able to find a model")
    raise typer.Exit()


def is_pipe(c: Tuple) -> bool:
    """Checks if | in c"""
    return "|" in c


def load_vocab(vocab: Path) -> List[List[str]]:
    """loads vocab from `vocab` and returns as list contaning lists of vocab"""
    with open(vocab, "r") as f:
        raw_vocab = [line.strip("\n") for line in f.readlines()]
        return list(
            map(
                list,
                (itertoolz.remove(is_pipe, (recipes.partitionby(is_pipe, raw_vocab)))),
            )
        )


@app.command()
def vocab(model: str = typer.Argument("latest"), show: bool = typer.Option(True)):
    """Prints out vocab for latest model"""
    if model != "latest":
        raise NotImplementedError
    model_dir = ensure_model_dir()
    if model_path := get_model(model_dir=model_dir):
        local_model = LocalModel(model_path)
        if local_model.vocab:
            vocab = load_vocab(local_model.vocab)
            if show:
                console.print(Markdown("# Model Vocab"))
                console.print(vocab)
            return vocab


def show_model_card(localmodel: LocalModel):
    """Shows model card for model"""
    with open(localmodel.modelcard, "r") as f:
        md = Markdown(f.read())
    console.print(md)


def hub_model_link(model_id: str) -> str:
    """Creates rich link for model card"""
    url = f"https://huggingface.co/{model_id}"
    return f"View [link={url}]model card[/link] for {model_id}!"


def create_markdown_model_card(model_id: str):
    """Creates rich Markdown wrapper for hub readme"""
    readme_url = hf_hub_url(model_id, filename="README.md")
    r = requests.get(readme_url)
    r.raise_for_status()
    return Markdown(r.text)


def create_metrics_tables(model_info: ModelInfo) -> List[Table]:
    """Creates a list of rich tables for metrics contained in `model_info`"""
    model_indexes = list(model_info.cardData["model-index"])
    metrics = []
    for model in model_indexes:
        for result in model["results"]:
            for metric in result["metrics"]:
                metrics.append(metric)
    for metric in metrics:
        table = Table()
        for name in metric.keys():
            table.add_column(name.title())
    metric_values = list(metric.values())
    rounded_metric_values = [
        round(item, ndigits=3) if isinstance(item, float) else item
        for item in metric_values
    ]
    table.add_row(*list(map(str, rounded_metric_values)))
    return [table]


if __name__ == "__main__":  # pragma: no cover
    app()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['flyswot']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=8,<10',
 'huggingface-hub>=0.2.1,<0.10.0',
 'numpy>=1.20,<2.0',
 'rich>=10.1,<13.0',
 'toolz>=0.11.1,<0.13.0',
 'transformers[torch]>=4.16.2,<5.0.0',
 'typer>=0.5.0',
 'typing_extensions>=3.10,<5.0']

entry_points = \
{'console_scripts': ['flyswot = flyswot.cli:main']}

setup_kwargs = {
    'name': 'flyswot',
    'version': '0.3.8',
    'description': 'flyswot',
    'long_description': '# flyswot\n\n[![PyPI](https://img.shields.io/pypi/v/flyswot.svg)][pypi_]\n[![Status](https://img.shields.io/pypi/status/flyswot.svg)][status]\n[![Python Version](https://img.shields.io/pypi/pyversions/flyswot)][python version]\n[![License](https://img.shields.io/pypi/l/flyswot)][license]\n\n[![Read the documentation at https://flyswot.readthedocs.io/](https://img.shields.io/readthedocs/flyswot/latest.svg?label=Read%20the%20Docs)][read the docs]\n[![Tests](https://github.com/davanstrien/flyswot/actions/workflows/tests.yml/badge.svg)](https://github.com/davanstrien/flyswot/actions/workflows/tests.yml)\n[![Codecov](https://codecov.io/gh/davanstrien/flyswot/branch/main/graph/badge.svg)][codecov]\n\n[![pre-commit](https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white)][pre-commit]\n[![Black](https://img.shields.io/badge/code%20style-black-000000.svg)][black]\n\n[pypi_]: https://pypi.org/project/flyswot/\n[status]: https://pypi.org/project/flyswot/\n[python version]: https://pypi.org/project/flyswot\n[license]: https://opensource.org/licenses/MIT\n[read the docs]: https://flyswot.readthedocs.io/\n[tests]: https://github.com/davanstrien/flyswot/actions?workflow=Tests\n[codecov]: https://app.codecov.io/gh/davanstrien/flyswot\n[pre-commit]: https://github.com/pre-commit/pre-commit\n[black]: https://github.com/psf/black\n\n![flyswot logo](https://raw.githubusercontent.com/davanstrien/flyswot/main/docs/_static/fly.png?token=ACEUI5KJ2HPO4ZGNTBX6OE3ARMXII)\n\n## Disclaimer 😬\n\n_flyswot_ is a work in progress. Things may not work and behaviour may change in the future!\n\n## tl;dr\n\n_flyswot_ is a Command Line Tool which allows you to run [Hugging Face Transformers](https://huggingface.co/docs/transformers/) [image classification models](https://huggingface.co/models?pipeline_tag=image-classification&sort=downloads) available via the [Hugging Face Hub](https://huggingface.co/models) 🤗 against a directory of images. It returns a CSV report containing the models predictions.\n\n```console\nflyswot predict directory image_directory csv_reports --model_id flyswot/convnext-tiny-224_flyswot\n```\n\n## Features\n\nCurrently _flyswot_ supports:\n\n- 🚀 automatic downloading of models from the Hugging Face Hub\n- 🔎 UNIX style search patterns for matching images to predict against\n- 📸 filtering by image extension\n- 📜 a CSV output report containing the paths to the input images, the predicted label and the models confidence for that prediction.\n- 📊 a summary \'report\' on the command line providing a high level summary of the predictions made\n\n[![asciicast](https://asciinema.org/a/492427.svg)](https://asciinema.org/a/492427)\n\n## Why?\n\nWhat is the point of this? Why not just write a Python script? This seems like a terrible idea...\n\n_flyswot_ was originally for a project with the [Heritage Made Digital](https://www.bl.uk/projects/heritage-made-digital) team at the [British Library](https://www.bl.uk). In this project we wanted to detect \'fake flysheets\'. We designed how _flyswot_ works with this particular use case in mind.\n\nThere are a few main reasons why we decided a command line tool was the best approaches to utilising the models we were developing.\n\n- The digitised images we are working with can be very large\n- The images we are working with are often subject to copyright\n- Inference speed isn\'t a big priority\n\nSince we\'re using computer vision for _assisting_ rather than _automation_ we felt a CLI was a useful interface for interacting with the models.\n\n## Installation\n\nYou can install _flyswot_ via [pip] from [PyPI]:\n\n```console\n$ pip install flyswot\n```\n\nThis will install the latest release version of _flyswot_\n\n## Detailed Installation Guide\n\n[Installation] provides a more detailed guide to installing _flyswot_. This more detailed guide is aimed at users of _flyswot_ who may be less familiar with Python.\n\n## Usage\n\nYou can see help for _flyswot_ using `flyswot --help`\n\n<!-- [[[cog\nimport cog\nfrom flyswot.cli import app\nfrom typer.testing import CliRunner\nrunner = CliRunner()\nresult = runner.invoke(app, ["--help"])\nhelp = result.output.replace("Usage: root", "Usage: flyswot")\ncog.out(\n    "```\\n{}\\n```".format(help)\n)\n]]] -->\n\n```\n\n Usage: flyswot [OPTIONS] COMMAND [ARGS]...\n\n╭─ Options ────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────╮\n│ --install-completion        [bash|zsh|fish|powershell|pwsh]  Install completion for the specified shell. [default: None]                                                 │\n│ --show-completion           [bash|zsh|fish|powershell|pwsh]  Show completion for the specified shell, to copy it or customize the installation. [default: None]          │\n│ --help                                                       Show this message and exit.                                                                                 │\n╰──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────╯\n╭─ Commands ───────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────╮\n│ model                      flyswot commands for interacting with models                                                                                                  │\n│ predict                    flyswot commands for making predictions                                                                                                       │\n╰──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────╯\n\n\n```\n\n<!-- [[[end]]] -->\n\n### Making predictions\n\nYou can get help for the prediction functionality for _flyswot_ as follows:\n\n<!-- [[[cog\nimport cog\nfrom flyswot.cli import app\nfrom typer.testing import CliRunner\nrunner = CliRunner()\nresult = runner.invoke(app,  ["predict", "directory", "--help"])\nhelp = result.output.replace("Usage: root", "Usage: flyswot")\ncog.out(\n    "```\\n{}\\n```".format(help)\n)\n]]] -->\n\n```\n\n Usage: flyswot predict directory [OPTIONS] DIRECTORY CSV_SAVE_DIR\n\n Predicts against all images stored under DIRECTORY which match PATTERN in the filename.\n By default searches for filenames containing \'fs\'.\n Creates a CSV report saved to `csv_save_dir`\n\n╭─ Arguments ──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────╮\n│ *    directory         PATH  Directory to start searching for images from [default: None] [required]                                                                     │\n│ *    csv_save_dir      PATH  Directory used to store the csv report [default: None] [required]                                                                           │\n╰──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────╯\n╭─ Options ────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────╮\n│ --model-id             TEXT     The model flyswot should use for making predictions [default: flyswot/convnext-tiny-224_flyswot]                                         │\n│ --pattern              TEXT     Pattern used to filter image filenames [default: None]                                                                                   │\n│ --bs                   INTEGER  Batch Size [default: 16]                                                                                                                 │\n│ --image-formats        TEXT     Image format(s) to check [default: .tif]                                                                                                 │\n│ --help                          Show this message and exit.                                                                                                              │\n╰──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────╯\n\n\n```\n\n<!-- [[[end]]] -->\n\nTo run predictions against a directory of images:\n\n```console\n$ flyswot predict directory manuscripts_folder .\n```\n\n- _flyswot_ will search inside the manuscripts_folder looking for image files.\n- By default it will look for files that contain `FS` in the filename since these are files which have been labelled as being "end flysheets" or "front flysheets"\n- Once it has found all the files labelled as `flysheet` it will then run a computer vision model against these images to see if they are labelled correctly i.e. if it is indeed a flysheet or something else.\n- _flyswot_ will save a csv report containing the paths to the image, the directory the image is stored in, the label, and the confidence for that prediction.\n\n#### Changing the model\n\nYou can also tell _flyswot_ to use a different image classification model via the `model-id` parameter. For example to use the [microsoft/dit-base-finetuned-rvlcdip](https://huggingface.co/microsoft/dit-base-finetuned-rvlcdip) model we could run:\n\n```console\nflyswot predict directory Documents/DS/hugit-cli/fs Desktop/ --model-id microsoft/dit-base-finetuned-rvlcdip\n\n```\n\nThis will download the latest available version of this model from the Hugging Face Hub and predict against the specified filenames. **Note** under the hood _flyswot_ uses the Hugging Face [transformers](https://huggingface.co/docs/transformers/) [pipelines](https://huggingface.co/docs/transformers/pipeline_tutorial) for inference. The model you specific must therefore be compatible with this pipeline.\n\n## Detailed Usage Guide\n\nThis section provides additional guidance on the usage of _flyswot_. This is primarily aimed at [HMD](https://www.bl.uk/projects/heritage-made-digital) users of _flyswot_.\n\n### How flyswot searches for images\n\n_flyswot_ is currently intended to identify images which have an incorrect label associated with them. In particular it is currently intended to identify "fake" flysheets. These images have `fs` as part of their filename so we can tell _flyswot_ to use this pattern in the filename to identify images which should be checked using the computer vision model. This can be changed if you also want to match other filename patterns.\n\nSince these images of concern will often be inside a directory structure _flyswot_ will look in sub-folders from the input folder for images which contain `fs` in the name. For example in the following folder structure:\n\n```console\nCollection/\n├─ item1/\n│  ├─ add_ms_9403_fbspi.tif\n│  ├─ add_ms_9403_fse001r.tif\n│  ├─ add_ms_9403_fse001v.tif\n├─ item2/\n│  ├─ sloane_ms_116_fblefr.tif\n│  ├─ sloane_ms_116_fbspi.tif\n│  ├─ sloane_ms_116_fse004r.tif\n```\n\nAll of the files which have `fs` in the filname will be check but files which don\'t contains `fs` such as `add_ms_9403_fbspi.tif` will be ignored since these aren\'t labelled as flysheets.\n\n### Running flyswot against a directory of images\n\nTo run _flyswot_ against a directory of images you need to give it the path to that directory/folder.\nThere are different ways you could do this. The following is suggested for people who are not very familiar (yet 😜) with terminal interfaces.\n\nIdentify the folder you want to _flyswot_ to check for "fake" flysheets. If you are using _flyswot_ for the first time it may make sense to choose a folder which doesn\'t contain a huge number of collection items so you don\'t have to wait to long for _flyswot_ to finish running. Once you have found a directory you want to predict against copy the path. This path should be the full path to the item.\n\nFor example something that looks like:\n\n```console\n\\\\ad\\collections\\hmd\\excitingcollection\\excitingsubcollection\\\n```\n\nThis will be the folder from which _flyswot_ starts looking.\n\nWhen you activated your conda environment in a terminal, you were likely \'inside\' your user directory. Since we need to specify a place for _flyswot_ to store the CSV report, we\'ll move to a better place to store that output; your `Desktop` folder. To do we can navigate using the command:\n\n```console\n$ chdir desktop\n```\n\nif you are using Mac, Linux or have GitBash installed you should instead run:\n\n```console\n$ cd Desktop\n```\n\nThis will take you to your Desktop. We\'ll now run _flyswot_. As with many other command line tools, _flyswot_ has commands and sub-commands. We are interested in the `predict` command. This includes two sub-commands: `predict-image` and `directory`. We will mostly want to predict directories. To do this we use the following approach. Since we only care about checking things with `fs` in the filename we can specify this as our `pattern`.\n\n```console\n$ flyswot predict directory input_directory output_directory --pattern fs\n```\n\nThe input directory is the folder containing our images and the output directory is where we want to save our CSV report. Using the folder we previously identified this would look like:\n\n```console\n$ flyswot predict directory "\\\\ad\\collections\\hmd\\excitingcollection\\excitingsubcollection\\" .\n```\n\nWe can use `.` to indicate we want the CSV report to be saved to the current directory (in this case the Deskop directory). Also notice that there are quotation marks `""` around the path. This is used to make sure that any spaces in the path are escaped.\n\nOnce you run this command you should see some progress reported by _flyswot_, including a progress bar that shows how many of the images _flyswot_ has predicted against.\n\nWhen _flyswot_ has finished you will have a CSV \'report\' which contains the path to the image, the predicted label and the confidence for that prediction.\n\n## License\n\nDistributed under the terms of the [MIT license],\n_flyswot_ is free and open source software.\n\n## Issues\n\nIf you encounter any problems,\nplease [file an issue] along with a detailed description.\n\n## Credits\n\nThis project was generated from [@cjolowicz]\'s [Hypermodern Python Cookiecutter] template.\n\n[@cjolowicz]: https://github.com/cjolowicz\n[cookiecutter]: https://github.com/audreyr/cookiecutter\n[mit license]: https://opensource.org/licenses/MIT\n[pypi]: https://pypi.org/\n[hypermodern python cookiecutter]: https://github.com/cjolowicz/cookiecutter-hypermodern-python\n[file an issue]: https://github.com/davanstrien/flyswot/issues\n[pip]: https://pip.pypa.io/\n[installation]: https://flyswot.readthedocs.io/en/latest/installation.html\n\n<!-- github-only -->\n\n[contributor guide]: https://github.com/davanstrien/flyswot/blob/main/CONTRIBUTING.md\n[background]: https://flyswot.readthedocs.io/en/latest/background.html\n',
    'author': 'Daniel van Strien',
    'author_email': 'daniel.van-strien@bl.uk',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/davanstrien/flyswot',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8.0,<4.0.0',
}


setup(**setup_kwargs)



"""Core inference functionality."""
from abc import ABC
from abc import abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Dict
from typing import Iterable
from typing import List
from typing import Union


class InferenceSession(ABC):
    """Abstract class for inference sessions"""

    @abstractmethod
    def __init__(self, model: Union[str, Path]):  # pragma: no cover
        """Inference Sessions should init from a model file and vocab"""
        pass

    @abstractmethod
    def predict_image(self, image: Path):  # pragma: no cover
        """Predict a single image"""
        pass

    @abstractmethod
    def predict_batch(
        self, model: Path, batch: Iterable[Path], bs: int
    ):  # pragma: no cover
        """Predict a batch"""
        pass


@dataclass
class ImagePredictionArgmaxItem:
    """Most confident predicted label for an item.

    Attributes:
        path: The Path to the image
        predicted_label: The predicted label i.e. the argmax value for the prediction tensor
        confidence: The confidence for `predicted_label` i.e. the max value for prediction tensor
    """

    path: Path
    predicted_label: str
    confidence: float

    def __post_init__(self) -> Union[Path, None]:
        """attempt to get absolute path"""
        try:
            self.path: Path = self.path.absolute()
        except AttributeError:
            pass


@dataclass
class MultiLabelImagePredictionItem:
    """Multiple predictions for a single image"""

    path: Path
    predictions: List[Dict[float, str]]

    def _get_top_labels(self) -> List[str]:
        """Get top labels"""
        top_labels = []
        for prediction in self.predictions:
            top_label = sorted(prediction.items(), reverse=True)[0][1]
            top_labels.append(top_label)
        return top_labels

    # Post init that gets top prediction label from predictions
    def __post_init__(self):
        """Get top prediction label"""
        self.predicted_labels = self._get_top_labels()


@dataclass
class PredictionBatch:
    """Container for ImagePredictionItems"""

    batch: List[ImagePredictionArgmaxItem]

    def __post_init__(self):
        """Returns a iterable of all predicted labels in batch"""
        self.batch_labels: Iterable[str] = (item.predicted_label for item in self.batch)


@dataclass
class MultiPredictionBatch:
    """Container for MultiLabelImagePredictionItems"""

    batch: List[MultiLabelImagePredictionItem]

    def _get_predicted_labels(self) -> Iterable:
        """Returns a iterable of all predicted labels in batch"""
        return (item.predicted_labels for item in self.batch)

    def __post_init__(self):
        """Returns a iterable of lists containing all predicted labels in batch"""
        # self.batch_labels: Iterable = (
        #     list(itertoolz.pluck(0, pred))
        #     for pred in zip(*[o.predictions for o in self.batch])
        # )
        self.batch_labels = self._get_predicted_labels()



"""Core functionality."""
import mimetypes
import time
from pathlib import Path
from typing import Any
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple

from toolz import itertoolz  # type: ignore

from flyswot.console import console


image_extensions: Set[str] = {
    k for k, v in mimetypes.types_map.items() if v.startswith("image/")
}


def count_files_with_ext(directory: Path, ext: str) -> int:
    """Counts files matching extension in directory"""
    return itertoolz.count(Path(directory).rglob(f"*{ext}"))


def create_file_search_message(
    directory: Path, pattern: Optional[str], ext: Optional[str]
) -> str:
    """Creates a message to show pattern and ext used for search"""
    if pattern and ext:
        message = f"Searching for files in {directory} matching {pattern} with extension {ext}"
    if pattern and not ext:
        message = f"Searching for image files in {directory} matching {pattern}"
    if not pattern and ext:
        message = f"Searching for all image files in {directory} with extension {ext}"
    if not pattern and not ext:
        message = f"Searching for all image files in {directory}"
    return message


def get_image_files_from_pattern(
    directory: Path, pattern: Optional[str], ext: Optional[str]
) -> Iterator[Path]:
    """yield image files from `directory` matching pattern `str` with `ext`"""
    message = create_file_search_message(directory, pattern, ext)
    with console.status(message, spinner="dots"):
        time.sleep(1)
        if pattern:
            yield from Path(directory).rglob(f"**/*{pattern}*{ext}")
            console.log("Search files complete...")
        if pattern and not ext:
            yield from Path(directory).rglob(f"*{pattern}*")
        if not pattern and ext:
            yield from Path(directory).rglob(f"*{ext}")
        if not pattern and not ext:
            match_files = Path(directory).rglob("*")
            for file in match_files:
                if file.suffix in image_extensions:
                    yield file


def filter_to_preferred_ext(files: Iterable[Path], exts: List[str]) -> Iterable[Path]:
    """filter_to_preferred_ext"""
    files = list(files)
    files_without_ext = (
        file.with_suffix("") for file in files if not file.name.startswith(".")
    )
    file_names_without_ext = (file.name for file in files_without_ext)
    unique = set(itertoolz.unique(file_names_without_ext))
    for file in files:
        file_without_suffix = file.with_suffix("")
        for ext in exts:
            if file_without_suffix.name in unique:
                if file.with_suffix(ext).is_file():
                    yield file.with_suffix(ext)
                else:
                    yield file
                unique.discard(file_without_suffix.name)


def signal_last(it: Iterable[Any]) -> Iterable[Tuple[bool, Any]]:
    """returns original iterator and iterator yield false until last item in iterator"""
    if not it:
        raise ValueError
    iterable = iter(it)
    ret_var = next(iterable)
    for val in iterable:
        yield False, ret_var
        ret_var = val
    yield True, ret_var
